use std::{collections::LinkedList, io::Read};

pub mod tcp;

pub mod entities {
    include!(concat!(env!("OUT_DIR"), "/messages.rs"));

    #[cfg(test)]
    mod tests {
        use super::*;
        use prost::Message;
        use std::io::Cursor;

        #[test]
        fn serialize_deserialize() {
            let serialized = CpuRequest { value: Some(1.0) };
            let vec = serialized.encode_to_vec();
            let deserialized = CpuRequest::decode(Cursor::new(vec)).unwrap();

            assert_eq!(serialized.value, deserialized.value);
        }
    }
}

pub mod routers {
    pub struct RequestRouter;
    impl RequestRouter {}
}

pub mod decoders {
    use std::io::Cursor;

    use prost::Message;

    use crate::entities::Request;

    pub struct RequestDecoder;

    impl RequestDecoder {
        pub fn decode(&self, buffer: Vec<u8>) -> Request {
            Request::decode(Cursor::new(buffer)).unwrap()
        }
    }
}

enum MessageKind {
    CPU,
    RAM,
    ProcessList,
}

pub struct VecByteBuffer {
    buffer: LinkedList<Vec<u8>>,
}

impl VecByteBuffer {
    fn push(&mut self, slice: &[u8], len: usize) {
        self.buffer.push_back(slice[..len].to_vec());
    }

    fn merge(&self) -> Vec<u8> {
        let mut merged = vec![];

        for partial in &self.buffer {
            for byte in partial {
                merged.push(*byte);
            }
        }

        merged
    }

    pub fn new() -> Self {
        VecByteBuffer {
            buffer: LinkedList::new(),
        }
    }
}

struct BufferedByteStream {
    buffer: Vec<u8>,
    start_ptr: usize,
}

impl BufferedByteStream {
    pub fn new(buffer: Vec<u8>) -> BufferedByteStream {
        BufferedByteStream {
            buffer,
            start_ptr: 0,
        }
    }
}

impl Read for BufferedByteStream {
    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
        let mut start = self.start_ptr;
        let boundary = start + buf.len();

        let end = if boundary >= self.buffer.len() {
            self.buffer.len()
        } else {
            boundary
        };

        let mut bytes_read = 0;
        while start < end {
            buf[bytes_read] = self.buffer[start];

            bytes_read += 1;
            start += 1;
        }

        self.start_ptr = boundary;

        Ok(bytes_read)
    }
}

#[cfg(test)]
mod tests {
    mod buffered_byte_stream_tests {
        use std::io::Read;

        use crate::BufferedByteStream;

        #[test]
        fn can_read_string_to_bytes_correctly() {
            let mut byte_stream = BufferedByteStream::new("hello".as_bytes().to_vec());

            let mut expected_buffer = [0u8; 2];
            let read_size = byte_stream.read(&mut expected_buffer).unwrap();

            assert_eq!(expected_buffer, [b'h', b'e']);
            assert_eq!(2, read_size);

            let mut expected_buffer = [0u8; 2];
            let read_size = byte_stream.read(&mut expected_buffer).unwrap();
            assert_eq!(expected_buffer, [b'l', b'l']);
            assert_eq!(2, read_size);

            let mut expected_buffer = [0u8; 2];
            let read_size = byte_stream.read(&mut expected_buffer).unwrap();
            assert_eq!(expected_buffer, [b'o', 0]);
            assert_eq!(1, read_size);
        }

        #[test]
        fn can_make_multiple_reads_buffer_smaller_than_stream() {
            let mut byte_stream = BufferedByteStream::new(vec![1, 2, 3, 4, 5]);

            let mut expected_buffer = [0u8; 3];
            let read_size = byte_stream.read(&mut expected_buffer).unwrap();

            assert_eq!(expected_buffer, [1, 2, 3]);
            assert_eq!(3, read_size);

            let mut expected_buffer = [0u8; 3];
            let read_size = byte_stream.read(&mut expected_buffer).unwrap();
            assert_eq!(2, read_size);

            assert_eq!(expected_buffer, [4, 5, 0]);
        }

        #[test]
        fn can_make_single_read_buffer_larger_than_stream() {
            let mut byte_stream = BufferedByteStream::new(vec![1, 2, 3, 4, 5]);

            let mut expected_buffer = [0u8; 6];

            let read_size = byte_stream.read(&mut expected_buffer).unwrap();

            assert_eq!(expected_buffer, [1, 2, 3, 4, 5, 0]);
            assert_eq!(5, read_size);
        }

        #[test]
        fn can_make_single_read_equal_length() {
            let mut byte_stream = BufferedByteStream::new(vec![1, 2, 3, 4, 5]);

            let mut expected_buffer = [0u8; 5];

            let read_size = byte_stream.read(&mut expected_buffer).unwrap();

            assert_eq!(expected_buffer, [1, 2, 3, 4, 5]);
            assert_eq!(5, read_size);
        }

        #[test]
        fn can_make_single_read_buffer_smaller_than_stream() {
            let mut byte_stream = BufferedByteStream::new(vec![1, 2, 3, 4, 5]);

            let mut expected_buffer = [0u8; 3];

            let read_size = byte_stream.read(&mut expected_buffer).unwrap();

            assert_eq!(expected_buffer, [1, 2, 3]);
            assert_eq!(3, read_size);
        }
    }

    mod vec_byte_buffer_tests {

        use crate::VecByteBuffer;

        #[test]
        fn push_bytes() {
            let mut buffer = VecByteBuffer::new();
            buffer.push(vec![1, 2, 3].as_ref(), 3);
            let merged = buffer.merge();
            assert_eq!(3, merged.len());

            buffer.push(vec![4, 5].as_ref(), 2);
            let merged = buffer.merge();
            assert_eq!(5, merged.len());
        }

        #[test]
        fn order_of_bytes_should_be_filo() {
            let mut buffer = VecByteBuffer::new();
            buffer.push(vec![1, 2, 3].as_ref(), 3);
            buffer.push(vec![4, 5].as_ref(), 2);

            let merged = buffer.merge();

            let mut index = 1;
            for byte in merged {
                assert_eq!(byte, index);
                index += 1;
            }
        }
    }
}
