use std::{
    io::{Read, Write},
    net::{TcpListener, TcpStream},
    str,
    sync::atomic::AtomicBool,
};

use crate::{decoders::RequestDecoder, routers::RequestRouter, VecByteBuffer};

struct TcpReader {
    temp_buffer_len: usize,
}

impl TcpReader {
    pub fn new(temp_buffer_len: usize) -> TcpReader {
        TcpReader { temp_buffer_len }
    }

    pub fn read<T: Read>(&self, stream: &mut T) -> Vec<u8> {
        let mut vec_byte_buffer = VecByteBuffer::new();

        loop {
            let mut buffer = vec![0u8; self.temp_buffer_len];
            let bytes_len = stream.read(buffer.as_mut_slice()).unwrap();

            // we are done reading, because client closed connection,
            // according to https://doc.rust-lang.org/stable/std/io/trait.Read.html#tymethod.read
            if bytes_len == 0 {
                break;
            }

            vec_byte_buffer.push(&buffer, bytes_len);
        }

        vec_byte_buffer.merge()
    }
}

pub struct StatusServer {
    stop: AtomicBool,
    reader: TcpReader,
    decoder: RequestDecoder,
    router: RequestRouter,
    pub port: i32,
    pub ip: String,
}

impl StatusServer {
    pub fn new(port: i32, ip: &str) -> StatusServer {
        StatusServer {
            ip: String::from(ip),
            reader: TcpReader::new(128),
            decoder: RequestDecoder,
            router: RequestRouter,
            port,
            stop: AtomicBool::new(false),
        }
    }

    pub fn start(&self) {
        let listener = TcpListener::bind(format!("{}:{}", self.ip, self.port)).unwrap();

        println!("starting server");
        for stream in listener.incoming() {
            if !self.stop.load(std::sync::atomic::Ordering::Relaxed) {
                println!("got stream");
                let mut stream = stream.unwrap();
                let data = self.reader.read(&mut stream);
                let request = self.decoder.decode(data);
            } else {
                println!("break");
                break;
            }
        }
    }

    pub fn stop(&self) {
        self.stop.store(true, std::sync::atomic::Ordering::Relaxed);
        println!("stop");

        let result = TcpStream::connect(format!("{}:{}", self.ip, self.port));

        let mut stream = match result {
            Ok(stream) => stream,
            Err(error) => {
                println!("error stopping {}", error);
                /* the server could be closed already because another request has activated the stop => do nothing */
                return;
            }
        };

        // we write an empty set to trigger the incoming loop that checks the stop bool that
        // has been set to true
        stream.write(b"");
        stream.flush();
        stream.shutdown(std::net::Shutdown::Both);
    }
}

#[cfg(test)]
mod tests {
    use std::io::Read;

    use crate::BufferedByteStream;

    use super::TcpReader;

    #[test]
    fn read_should_break_on_empty() {
        let reader = TcpReader::new(2);
        let mut mock_stream = MockTcpStream::new("hello".as_bytes().to_vec());
        let data = reader.read(&mut mock_stream);
        let string = String::from_utf8(data).unwrap();
        assert_eq!(string, "hello");
    }

    struct MockTcpStream {
        buffer: BufferedByteStream,
    }

    impl MockTcpStream {
        fn new(buf: Vec<u8>) -> MockTcpStream {
            MockTcpStream {
                buffer: BufferedByteStream::new(buf),
            }
        }
    }

    impl Read for MockTcpStream {
        fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
            self.buffer.read(buf)
        }
    }
}
